#include <cstring>
#include <cmath>

#include "colors.hh"
#include "strfun.hh"

struct ColorInfo { const char*name; unsigned color; };

static bool FindColorInfo(const ColorInfo&a, const ColorInfo& b)
{
    return std::strcmp(a.name, b.name) < 0;
}

unsigned TranslateColorNameToRGB(const std::string& name)
{
    #define RGB(r,g,b) (((r)<<16) + ((g)<<8) + ((b)))
    std::string nameu = UCase(name);
    
    //fprintf(stderr, "Color name for '%s'...\n", nameu.c_str());
    
    if(nameu.size() == 4 && nameu[0] == '#')
    {
        unsigned val = strtol(nameu.c_str()+1, 0, 16);
        unsigned r = (val&0xF00)>>8;  r*=0x11;
        unsigned g = (val&0xF0)>>4;   g*=0x11;
        unsigned b = (val&0xF)>>0;    b*=0x11;
        return RGB(r,g,b);
    }
    if(nameu.size()==7 && nameu[0] == '#')
    {
        return std::strtol(nameu.c_str()+1, 0, 16);
    }
    static const ColorInfo colors[] =
    {
#include "colors.inc"
    };
    
    ColorInfo dummy; dummy.name = nameu.c_str();
    const ColorInfo* endcolors = colors+ (sizeof(colors)/sizeof(*colors));
    
    const ColorInfo* color = std::lower_bound( colors, endcolors, dummy, FindColorInfo);
    if(color!=endcolors) if(nameu != color->name) color=endcolors;
    
    /* fprintf(stderr, "Color '%s': %X (%s)\n",
        dummy.name,
        color==endcolors ? 0xC0C0C0 : color->color,
        color==endcolors ? "not found" : "found"
    );*/
    
    if(color != endcolors)
        return color->color;
    
    if(nameu.size() == 6)
        return std::strtol(nameu.c_str(), 0, 16);
    
    return 0xC0C0C0;
    #undef RGB
}

double RgbDiff(unsigned rgb1, unsigned rgb2)
{
    unsigned R1 = ((rgb1>>16)&0xFF), R2 = ((rgb2>>16)&0xFF);
    unsigned G1 = ((rgb1>> 8)&0xFF), G2 = ((rgb2>> 8)&0xFF);
    unsigned B1 = ((rgb1>> 0)&0xFF), B2 = ((rgb2>> 0)&0xFF);
    int Y1=R1,U1=G1,V1=B1, Y2=R2,U2=G2,V2=B2;
    /*
    double Y1 = (( ( 66 * R1 + 129 * G1 + 25 * B1 + 128) / 256.0) + 16);
    double U1 = ( ( -38 * R1 - 74 * G1 + 112 * B1 + 128) / 256.0) + 128;
    double V1 = ( ( 112 * R1 - 94 * G1 - 18 * B1 + 128) / 256.0) + 128;
    double Y2 = (( ( 66 * R2 + 129 * G2 + 25 * B2 + 128) / 256.0) + 16);
    double U2 = ( ( -38 * R2 - 74 * G2 + 112 * B2 + 128) / 256.0) + 128;
    double V2 = ( ( 112 * R2 - 94 * G2 - 18 * B2 + 128) / 256.0) + 128;
    */
    return std::pow((Y1-Y2),2.0)
         + std::pow((U1-U2),2.0)
         + std::pow((V1-V2),2.0);
}

unsigned CreateXtermColor(const unsigned rgb)
{
    #define sixtable(n)  (n ? 55+(n)*40 : 0)
    #define graytable(n) (8 + 10*(n))  /* n=0..23 */
    unsigned r = (rgb>>16)&0xFF, g = (rgb>>8)&0xFF, b = (rgb>>0)&0xFF;
    
    // Try six-algorithm
    unsigned r6 = r*6/256, g6=g*6/256, b6=b*6/256;
    unsigned six_result = (sixtable(r6)<<16) + (sixtable(g6)<<8) + (sixtable(b6));
    
    // Try grayscale-algorithm
    unsigned gray_value  = (r*30+g*60+b*11)/101; // approximate Y formula
    gray_value = (gray_value<8) ? 0 : (gray_value>0xEE ? 23 : (gray_value-8+5)/10);
    unsigned gray_result = graytable(gray_value);
    
    if(RgbDiff(gray_result,rgb) < RgbDiff(six_result,rgb))
    {
        return 232 + gray_value;
    }
    // Of the first 16 ansi colours, only black we can trust
    // that it's some certain value. The others vary wildly.
    if(six_result == 0) return 0; // black
    return 16 + b6 + g6*6 + r6*6*6;
}
